"""OAT algorithm configuration interface for unified planning."""

from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
from ConfigSpace.conditions import (
    AndConjunction
)
import random
import sys
import os
import subprocess
import importlib

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from AC_interface import GenericACInterface
from utils.pcs_transform import transform_pcs


class OATInterface(GenericACInterface):
    """
    OAT AC interface.

    OAT does not handle forbidden parameter value combinations.
    OAT can handle multiple parent and 1 child conditionals,
    but not one parent multiple children conditionals.
    We naively just take the first one in the list.
    OAT does not support conditionals that are conditional.
    We leave them out naively.
    OAT does not support conditionals with value ranges.
    We naively only use the first value.

    Note: Although this is suboptimal, invalid configurations will
    lead to crash or bad results such that OAT will rate them
    as subpar.
    
    """

    def __init__(self):
        """Initialize OAT interface."""
        GenericACInterface.__init__(self)

        if importlib.metadata.version('ConfigSpace') != '0.6.1':
            print('Installing ConfigSpace 0.6.1')
            subprocess.check_call([sys.executable, "-m", "pip", "install",
                                   "ConfigSpace==0.6.1"],
                                  stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)

            os.execv(sys.executable, ['python'] + sys.argv)

    def transform_conf_from_ac(self, engine, configuration, plantype):
        """
        Transform a configuration to the UP engine format.

        :param engine: The name of the engine.
        :type engine: str
        :param configuration: Parameter names with values.
        :type configuration: dict

        :return: The transformed configuration.
        :rtype: dict
        """
        config = transform_pcs(engine, configuration)
        if engine in ('lpg', 'lpg-anytime'):
            del_list = []
            add_list = []
            for pname, pvalue in config.items():
                if pname in self.engine_param_types[engine]:
                    if self.engine_param_types[engine][pname] == 'FLAGS':
                        del_list.append(pname)
                        flag_pname = pname + '=' + config[pname]
                        add_list.append(flag_pname)
                    elif self.engine_param_types[engine][pname] == 'FLAG':
                        if config[pname] == '1':
                            config[pname] = ''
                        else:
                            del_list.append(pname)

            for dl in del_list:
                del config[dl]

            for al in add_list:
                config[al] = ''

        elif engine == 'fast-downward':
            if len(config) == 1:
                pass
            else:
                if config['fast_downward_search_config'] == 'astar':
                     
                    if config['astar_h'] == 'merge_and_shrink':
                        search_option = config['fast_downward_search_config'] + '('
                        greedy = config['greedy']
                        before_shrinking = config['before_shrinking']
                        before_merging = config['before_merging']
                        if before_shrinking == 'false' and before_merging == 'false':
                            before_merging = 'true'
                        max_states = int(float(config['max_states']))
                        threshold_before_merge = config['threshold_before_merge']

                        search_option += 'merge_and_shrink(merge_strategy='
                        search_option += 'merge_precomputed(merge_tree=linear(variable_order='
                        search_option += 'reverse_level)),shrink_strategy='
                        search_option += f'shrink_bisimulation(greedy={greedy}),'
                        search_option += 'label_reduction='
                        search_option += f'exact(before_shrinking={before_shrinking},'
                        search_option += f'before_merging={before_merging}),'
                        search_option += f'max_states={max_states}'

                        if max_states == 'true': 
                            threshold_before_merge += ',threshold_before_merge=1))'
                        else:
                            search_option += '))'
                    else:
                        search_option = config['fast_downward_search_config'] + '('
                        search_option += config['astar_h'] + '())'
                elif config['fast_downward_search_config'] in ('lazy_greedy', 'eager_greedy'):
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}']
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                        if idx > len(h_name):
                            idxxx = idx - skip
                        else:
                            idxxx = idx
                        h_def += f'let(h{h_name[idxxx - 1][:2]},{h}(),'
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += 'h' + hn[:2]
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if 'h' + h_name[h_idx][:2] not in used not in used:
                                prefs += 'h' + h_name[h_idx][:2]
                        if idx < len(h_name) and 'h' + h_name[h_idx][:2] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if 'h' + h_name[h_idx][:2] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append('h' + h_name[h_idx][:2])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'
                    for _ in range(len(h_name) - 1):
                        search_option += ')'

                elif config['fast_downward_search_config'] == 'lazy_wastar':
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}']
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                        if idx > len(h_name):
                            idxxx = idx - skip
                        else:
                            idxxx = idx
                        h_def += f'let(h{h_name[idxxx - 1][:2]},{h}(),'
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += 'h' + hn[:2]
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    w = int(float(config['w']))
                    search_option += f'w={w},'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1    
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if 'h' + h_name[h_idx][:2] not in used not in used:
                                prefs += 'h' + h_name[h_idx][:2]
                        if idx < len(h_name) and 'h' + h_name[h_idx][:2] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if 'h' + h_name[h_idx][:2] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append('h' + h_name[h_idx][:2])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'
                    for _ in range(len(h_name) - 1):
                        search_option += ')'

                elif config['fast_downward_search_config'] == 'eager':
                    h = config['h_1']
                    h_name = h
                    search_option = ''
                    h_def = f'let(h{h_name[:2]},{h}(),'
                    weight = int(float(config['weight']))
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(single(sum([g(),weight(' + f'h{h_name[:2]}'
                    search_option += f',{weight})])),'
                    search_option += f'preferred=[h{h_name[:2]}],'
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'

                if plantype == 'OneshotPlanner':
                    config = {'fast_downward_search_config': search_option}

                elif plantype == 'AnytimePlanner':
                    config = {'fast_downward_anytime_search_config': search_option}

        elif engine == 'symk':

            if len(config) == 1:
                pass
            else:
                if config['fast_downward_search_config'] == 'astar':
                     
                    if config['astar_h'] == 'merge_and_shrink':
                        search_option = config['fast_downward_search_config'] + '('
                        greedy = config['greedy']
                        before_shrinking = config['before_shrinking']
                        before_merging = config['before_merging']
                        if before_shrinking == 'false' and before_merging == 'false':
                            before_merging = 'true'
                        max_states = int(float(config['max_states']))
                        threshold_before_merge = config['threshold_before_merge']

                        search_option += 'merge_and_shrink(merge_strategy='
                        search_option += 'merge_precomputed(merge_tree=linear(variable_order='
                        search_option += 'reverse_level)),shrink_strategy='
                        search_option += f'shrink_bisimulation(greedy={greedy}),'
                        search_option += 'label_reduction='
                        search_option += f'exact(before_shrinking={before_shrinking},'
                        search_option += f'before_merging={before_merging}),'
                        search_option += f'max_states={max_states}'

                        if max_states == 'true': 
                            threshold_before_merge += ',threshold_before_merge=1))'
                        else:
                            search_option += '))'
                    else:
                        search_option = config['fast_downward_search_config'] + '('
                        search_option += config['astar_h'] + '())'

                elif config['fast_downward_search_config'] in ('lazy_greedy', 'eager_greedy'):
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}'] + '()'
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += hn
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if h_name[h_idx] not in used:
                                prefs += h_name[h_idx]
                        if idx < len(h_name) and h_name[h_idx] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if h_name[h_idx] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append(h_name[h_idx])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                elif config['fast_downward_search_config'] == 'lazy_wastar':
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}'] + '()'
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += hn
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    w = int(float(config['w']))
                    search_option += f'w={w},'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1    
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if h_name[h_idx] not in used not in used:
                                prefs += h_name[h_idx]
                        if idx < len(h_name) and h_name[h_idx] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if h_name[h_idx] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append(h_name[h_idx])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                elif config['fast_downward_search_config'] == 'eager':
                    h = config['h_1'] + '()'
                    h_name = h
                    search_option = ''
                    weight = int(float(config['weight']))
                    search_option += config['fast_downward_search_config']
                    search_option += '(single(sum([g(),weight(' + f'{h_name}'
                    search_option += f',{weight})])),'
                    search_option += f'preferred=[{h_name}],'
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                if plantype == 'OneshotPlanner':
                    config = {'symk_search_config': search_option}

                elif plantype == 'AnytimePlanner':
                    config = {'symk_anytime_search_config': search_option}

        elif engine in ('tamer', 'pyperplan'):
            config = {}
            for param in \
                    self.engine_param_spaces[engine].get_hyperparameters():
                if isinstance(param, UniformFloatHyperparameter):
                    config[param.name] = float(configuration[param.name])
                elif isinstance(param, UniformIntegerHyperparameter):
                    config[param.name] = int(configuration[param.name])
                else:
                    config[param.name] = configuration[param.name]

        elif engine in ('enhsp', 'enhsp-any'):
            if not isinstance(configuration, str) and 'params' not in configuration:
                config = ''
                for name, value in configuration.items():
                    config += '-' + name[0] + ' ' + value + ' '
                config = {'params': config}
            else:
                config = configuration
 
        return config

    def get_ps_oat(self, param_space):
        """
        Generate the OAT parameter tree in XML format.

        OAT does not handle forbidden parameter value combinations.
        OAT can handle multiple parent and 1 child conditionals,
        but not one parent multiple children conditionals.
        We naively just take the first one in the list.
        OAT does not support conditionals that are conditional.
        We leave them out naively.
        OAT does not support conditionals with value ranges.
        We naively only use the first value.

        Note: Although this is suboptimal, invalid configurations will
        lead to crash or bad results such that OAT will rate them
        as subpar.

        :param param_space: ConfigSpace object.
        :type param_space: ConfigSpace.ConfigurationSpace

        :return: OAT parameter tree in XML format.
        :rtype: str

        """

        param_file = '<?xml version="1.0" encoding="utf-8" ?>\n'
        param_file += \
            '<node xsi:type="and" xsi:noNamespaceSchemaLocation="' + \
            '../parameterTree.xsd" xmlns:xsi="http://www.w3.org/' + \
            '2001/XMLSchema-instance">\n'

        hyperparameters = param_space.get_hyperparameters()
        to_set = []
        for hp in hyperparameters:
            to_set.append(hp.name)

        conditions = param_space.get_conditions()

        parents = {}
        for cond in conditions:
            if not isinstance(cond, AndConjunction):
                if cond.parent.name not in parents:
                    parents[cond.parent.name] = \
                        {cond.child.name: [cond.value, cond.child]}
                else:
                    parents[cond.parent.name][cond.child.name] = \
                        [cond.value, cond.child]
            else:
                for c in cond.components:
                    if c.parent.name not in parents:
                        parents[c.parent.name] = \
                            {c.child.name: [c.value, c.child]}
                    else:
                        parents[c.parent.name][c.child.name] = \
                            [c.value, c.child]

        def set_conditionals(children, param_file, to_set, parents, tab=''):
            """
            Set conditional relations between parameters.

            :param children: Child parameters.
            :type children: dict
            :param param_file: OAT parameter tree to be saved in XML.
            :type param_file: str
            :param to_set: Parameter names to still be included.
            :type to_set: list
            :param parents: Parent parameters.
            :type parents: dict
            :param tab: Indicates depth of tree (\t).
            :type tab: str

            :return: Updated OAT parameter tree and parameter names to still be included.
            :rtype: tuple
            """
            for child, value in children.items():
                if child in to_set:
                    if isinstance(value[0], list):
                        value[0] = value[0][0]
                    param_file += f'{tab}\t\t<choice>\n'
                    param_file += f'{tab}\t\t\t<string>{value[0]}</string>\n'
                    param_file += \
                        f'{tab}\t\t\t<child xsi:type="value" id="{child}">\n'
                    if isinstance(value[1], CategoricalHyperparameter):
                        choices = ''
                        for c in value[1].choices:
                            choices += f'{c} '
                        param_file += f'{tab}\t\t\t\t<domain xsi:type=' + \
                            '"categorical" strings="{choices[:-1]}" ' + \
                            'defaultIndexOrValue="{value[1].choices.' + \
                            'index(value[1].default_value)}"/>\n'
                        param_file += f'{tab}\t\t\t</child>\n'
                        param_file += f'{tab}\t\t</choice>\n'
                    elif isinstance(value[1], UniformIntegerHyperparameter):
                        if value[1].lower < -2147483647:
                            lower = -2147483647
                        else:
                            lower = value[1].lower
                        if value[1].upper > 2147483647:
                            upper = 2147483647
                        else:
                            upper = value[1].upper
                        param_file += f'{tab}\t\t\t\t<domain xsi:type=' + \
                            f'"discrete" start="{lower}" end="{upper}" ' + \
                            'defaultIndexOrValue=' + \
                            f'"{value[1].default_value}"/>\n'
                        param_file += f'{tab}\t\t\t</child>\n'
                        param_file += f'{tab}\t\t</choice>\n'
                    elif isinstance(value[1], UniformFloatHyperparameter):
                        param_file += f'{tab}\t\t\t\t<domain xsi:type=' + \
                            f'"continuous" start="{value[1].lower}" end=' + \
                            f'"{value[1].upper}" defaultIndexOrValue=' + \
                            f'"{value[1].default_value}"/>\n'
                        param_file += f'{tab}\t\t\t</child>\n'
                        param_file += f'{tab}\t\t</choice>\n'

                    to_set.remove(child)              

            return param_file, to_set

        for param in hyperparameters:
            if param.name in to_set:
                if param.name in parents and \
                        parents[param.name].keys() in to_set:
                    param_file += f'\t<node xsi:type="or" id="{param.name}">\n'
                    if isinstance(param, CategoricalHyperparameter):
                        choices = ''
                        for c in param.choices:
                            choices += f'{c} '
                        param_file += \
                            '\t\t<domain xsi:type="categorical" strings=' + \
                            f'"{choices[:-1]}" defaultIndexOrValue=' + \
                            f'"{param.choices.index(param.default_value)}"/>\n'

                        children = parents[param.name]
                        param_file, to_set = \
                            set_conditionals(children, param_file, to_set,
                                             parents)
                        param_file += '\t</node>\n'                       

                    elif isinstance(param, UniformIntegerHyperparameter):
                        if param.lower < -2147483647:
                            lower = -2147483647
                        else:
                            lower = param.lower
                        if param.upper > 2147483647:
                            upper = 2147483647
                        else:
                            upper = param.upper
                        param_file += \
                            '  <domain xsi:type="discrete" start=' + \
                            f'"{lower}" end="{upper}" defaultIndexOrValue=' + \
                            f'"{param.default_value}"/>\n'

                        children = parents[param.name]
                        param_file, to_set = \
                            set_conditionals(children, param_file, to_set,
                                             parents)
                        param_file += '\t</node>\n'

                    elif isinstance(param, UniformFloatHyperparameter):
                        param_file += \
                            '\t\t<domain xsi:type="continuous" start=' + \
                            f'"{param.lower}" end="{param.upper}" ' + \
                            f'defaultIndexOrValue="{param.default_value}"/>\n'
                        
                        children = parents[param.name]
                        param_file, to_set = \
                            set_conditionals(children, param_file, to_set,
                                             parents)
                        param_file += '\t</node>\n'
                else:
                    param_file += \
                        f'\t<node xsi:type="value" id="{param.name}">\n'
                    if isinstance(param, CategoricalHyperparameter):
                        choices = ''
                        for c in param.choices:
                            choices += f'{c} '
                        param_file += \
                            '\t\t<domain xsi:type="categorical" strings=' + \
                            f'"{choices[:-1]}" defaultIndexOrValue=' + \
                            f'"{param.choices.index(param.default_value)}"/>\n'
                        param_file += '\t</node>\n'
                    elif isinstance(param, UniformIntegerHyperparameter):
                        if param.lower < -2147483647:
                            lower = -2147483647
                        else:
                            lower = param.lower
                        if param.upper > 2147483647:
                            upper = 2147483647
                        else:
                            upper = param.upper
                        param_file += \
                            '\t\t<domain xsi:type="discrete" start=' + \
                            f'"{lower}" end="{upper}" defaultIndexOrValue=' + \
                            f'"{param.default_value}"/>\n'
                        param_file += '\t</node>\n'
                    elif isinstance(param, UniformFloatHyperparameter):
                        param_file += '\t\t<domain xsi:type="continuous"' + \
                            f' start="{param.lower}" end="{param.upper}"' + \
                            f' defaultIndexOrValue="{param.default_value}"/>\n'
                        param_file += '\t</node>\n'

                to_set.remove(param.name)

        param_file += '</node>\n'

        return param_file



import dill
from unified_planning.io import PDDLReader
import sys
import os

# Read problem instance path
with open(sys.argv[1], 'r') as f:
    instance = f.read()

# Construct parameter dict
params = {}
for i, arg in enumerate(sys.argv):
    sys.argv[i] = arg.replace('{', '').replace('}', '')

for arg in sys.argv[2::2]:
    params[arg[1:]] = sys.argv[sys.argv.index(arg) + 1]
    
# Make sure modules are accesible
path = os.getcwd().rsplit('up_ac', 2)[0]
if path[-1] != "/":
    path += "/"
path += 'up_ac'
sys.path.append(r"{}".format(path))

# Load planner_feedback function
feedback = dill.load(open(f'{path}/OAT/feedback.pkl', 'rb'))
feedback = feedback(params, instance, PDDLReader())

# Print feedback for OAT to parse
print('\n\n', feedback)



"""Generic algorithm configuration interface for unified planning."""
import unified_planning
from unified_planning.environment import get_environment
from unified_planning.shortcuts import *

from tarski.io import PDDLReader as treader
import logging
import traceback
import os
import sys

# make sure test can be run from anywhere
path = os.getcwd().rsplit('up-ac', 1)[0]
if path[-1] != "/":
    path += "/"
path += 'up-ac/up_ac'
if not os.path.isfile(sys.path[0] + '/configurators.py') and \
        'up-ac' in sys.path[0]:
    sys.path.append(sys.path[0].rsplit('up-ac', 1)[0] + 'up-ac')
    sys.path.append(sys.path[0].rsplit('up-ac', 1)[0] + 'up-ac/up_ac')
    sys.path.append(sys.path[0].rsplit('up-ac', 1)[0] + 'up-ac/utils')

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from utils.ac_feedback import qaul_feedback, runtime_feedback
from utils.timeout import linux_timeout
from utils.parametrizable_engines import parametrizable_engines


class GenericACInterface():
    """Generic AC interface."""

    def __init__(self):
        """Initialize generic interface."""
        self.environment = get_environment()
        self.available_engines = self.get_available_engines()
        self.engine_param_spaces = {}
        self.engine_param_types = {}
        self.treader = treader(raise_on_error=True)
        self.verbose = True

    def get_available_engines(self):
        """Get planning engines installed in up."""
        factory = unified_planning.engines.factory.Factory(self.environment)

        return factory.engines

    def compute_instance_features(self, domain, instance):
        """
        Compute instance features of a given PDDL instance.

        :param domain: PDDL string representing the problem domain.
        :type domain: str
        :param instance: PDDL string representing the problem instance.
        :type instance: str

        :returns: Computed instance features.
        :rtype: list
        """
        logging.getLogger().setLevel(logging.ERROR)
        try:
            # TODO catch duplicate errors in tarski
            features = []
            self.treader.parse_domain(domain)
            problem = self.treader.parse_instance(instance)
            lang = problem.language
            features.append(len(lang.predicates))
            features.append(len(lang.functions))
            features.append(len(lang.constants()))
            features.append(len(list(problem.actions)))
            features.append(features[1] / features[0])
            features.append(features[1] / features[2])
            features.append(features[1] / features[3])
            features.append(features[0] / features[2])
            features.append(features[0] / features[3])
            features.append(features[2] / features[3])
        except:
            features = [0 for _ in range(10)]

        return features

    def read_engine_pcs(self, engines, pcs_dir):
        """
        Read parameter configuration space (PCS) files for specified engines.

        :param engines: Names of the engines.
        :type engines: list of str
        :param pcs_dir: Path to the directory containing the PCS files.
        :type pcs_dir: str
        """

        if pcs_dir[-1] != '/':
            pcs_dir = pcs_dir + '/'

        for engine in engines:
            import ConfigSpace
            from utils.patches import patch_pcs
            ConfigSpace.read_and_write.pcs = \
                patch_pcs(ConfigSpace.read_and_write.pcs)

            with open(pcs_dir + engine + '.pcs', 'r') as f:
                self.engine_param_spaces[engine] = \
                    ConfigSpace.read_and_write.pcs.read(f)

            with open(pcs_dir + engine + '.pcs', 'r') as f:
                lines = f.readlines()
                self.engine_param_types[engine] = {}
                for line in lines:
                    if '# FLAGS #' in line:
                        self.engine_param_types[engine][
                            '-' + line.split(' ')[0]] = 'FLAGS'
                    elif '# FLAG' in line:
                        self.engine_param_types[engine][
                            '-' + line.split(' ')[0]] = 'FLAG'

    def get_feedback(self, engine, fbtype, result):
        """
        Get feedback from a planning engine after a run.

        :param engine: Name of the planning engine.
        :type engine: str
        :param fbtype: Type of feedback: 'quality' or 'runtime'.
        :type fbtype: str
        :param result: Planning result.
        :type result: object

        :returns: Feedback based on the specified feedback type.
        :rtype: object

        :raises ValueError: If an unsupported feedback type is provided.
        """
        if fbtype == 'quality':
            feedback = qaul_feedback(engine, result)
        if fbtype == 'runtime':
            feedback = runtime_feedback(engine, result)

        return feedback

    def compute_plan_cost(self, problem, res):
        """
        Compute plan cost based on up problem and result.

        :param problem: up problem
        :type problem: up.model.Problem
        :param res: Result provided by a planning engine
        :type res: up.engines.results.Result

        :returns: Cost of the plan
        :type: int
        """
        try:    
            if not problem._metrics:
                total_cost = len(res.plan.actions)

            else:
                total_cost = 0

                if problem._metrics[0].is_minimize_action_costs():
                    action_costs = {
                        k.name: v
                        for k, v in problem._metrics[0].costs.items()
                    }

                    action_costs = {
                        k: v.constant_value()
                        for k, v in action_costs.items()
                    }

                    # Iterate actions in plan
                    for action_instance in res.plan.actions:
                        action_name = action_instance.action.name
                        total_cost += (
                            action_costs[action_name]
                            if action_name in action_costs
                            else 0
                        )
                else:
                    cost_fluent = next(
                        (fl for fl in problem.fluents
                            if "total-cost" in fl.name.lower()
                            and fl.type.is_real_type()),
                        None
                    )
                    
                    if cost_fluent is None:
                        raise ValueError("""No total-cost fluent found in the
                            problem definition.""")

                    cost_fluent_fnode = None
                    for fnode in problem.initial_values.keys():
                        if fnode.is_fluent_exp() and fnode.fluent() \
                                == cost_fluent:
                            cost_fluent_fnode = fnode
                            break

                    if cost_fluent_fnode is None:
                        raise ValueError(f"""Could not find an instance of
                            fluent {cost_fluent.name} in the initial state.""")

                    initial_cost = \
                        problem.initial_values.get(cost_fluent_fnode, 0.0)

                    initial_values = {}
                    for key, val in problem.initial_values.items():
                        initial_values[str(key)] = val

                    total_cost = initial_cost.constant_value()

                    for action_instance in res.plan.actions:
                        action = action_instance.action
                        substitutions = \
                            dict(zip(action.parameters,
                                     action_instance.actual_parameters))

                        for effect in action.effects:
                            if effect.fluent.fluent() == cost_fluent:
                                # Substitute action param to effect expression
                                effect_value_expr = \
                                    effect.value.substitute(substitutions)
                                try:
                                    evaluated_value = \
                                        problem.initial_values.get(
                                            effect_value_expr,
                                            None).constant_value()
                                    if evaluated_value is None:
                                        evaluated_value = \
                                            effect_value_expr.constant_value()
                                except AssertionError:
                                    print(f"""Warning: Could not evaluate
                                        effect value {effect_value_expr},
                                        using 0.""")
                                    evaluated_value = 0

                                if effect.is_assignment():
                                    total_cost = evaluated_value
                                elif effect.is_increase():
                                    total_cost += evaluated_value
                                elif effect.is_decrease():
                                    total_cost -= evaluated_value

            return total_cost
        except Exception:
            traceback.print_exc()
            return None

    def run_engine_config(self, config, metric, engine, plantype, problem,
                          timelimit, gb_listener=None):
        """
        Execute a configured engine run.

        :param config: Configuration of the engine.
        :type config: dict
        :param metric: Metric for the evaluation: 'runtime' or 'quality'.
        :type metric: str
        :param engine: Name of the engine.
        :type engine: str
        :param plantype: Type of planning: 'OneshotPlanner' or 'AnytimePlanner'.
        :type plantype: str
        :param problem: Path to the problem instance.
        :type problem: str
        :param gb_listener: True if using a gray box approach (optional).
        :type gb_listener: bool

        :returns: Result from the configured engine run.
        :rtype: object

        :raises ValueError: If an unsupported planning type is provided.
        """
        if plantype == 'OneshotPlanner':
            config = self.transform_conf_from_ac(engine, config, plantype)
            if gb_listener is not None:
                with OneshotPlanner(name=engine,
                                    params=config,
                                    output_stream=gb_listener) as planner:
                    try:
                        planner = \
                            linux_timeout(planner, engine, timelimit, config)
                        if engine == 'tamer' or engine == 'pyperplan':
                            result = planner.solve(problem)
                        else:
                            result = planner.solve(problem, timeout=timelimit)
                        if (result.status ==
                                up.engines.PlanGenerationResultStatus.
                                SOLVED_SATISFICING or result.status ==
                                up.engines.PlanGenerationResultStatus.
                                SOLVED_OPTIMALLY):
                            if self.verbose:
                                print("Result found.\n")
                            if (engine == 'tamer' or engine == 'pyperplan'):
                                if metric == 'quality':
                                    feedback = self.compute_plan_cost(problem,
                                                                      result)
                                else:
                                    feedback = self.get_feedback(engine,
                                                                 metric,
                                                                 result)
                            else:
                                feedback = self.get_feedback(engine,
                                                             metric,
                                                             result)
                        elif result.status == \
                                up.engines.PlanGenerationResultStatus.\
                                UNSOLVABLE_PROVEN:
                            if self.verbose:
                                print("Problem proven unsolvable.\n")
                            feedback = 'unsolvable'
                        elif result.status == \
                                up.engines.PlanGenerationResultStatus.\
                                INTERNAL_ERROR:
                            if self.verbose:
                                print("Internal Error.\n")
                            feedback = None
                        else:
                            if self.verbose:
                                print("No plan found.\n")
                            feedback = None
                    except Exception as e:
                        if self.verbose:
                            print("No plan found.\n")
                            print('Exception:', e)
                        feedback = None
            else:
                
                with OneshotPlanner(name=engine,
                                    params=config) as planner:
                    try:
                        planner = \
                            linux_timeout(planner, engine, timelimit, config)
                        if engine == 'tamer' or engine == 'pyperplan':
                            result = planner.solve(problem)
                        else:
                            result = planner.solve(problem, timeout=timelimit)
                        if (result.status ==
                                up.engines.PlanGenerationResultStatus.
                                SOLVED_SATISFICING or result.status ==
                                up.engines.PlanGenerationResultStatus.
                                SOLVED_OPTIMALLY):
                            if self.verbose:
                                print("Result found.\n")
                            if (engine == 'tamer' or engine == 'pyperplan'):
                                if metric == 'quality':
                                    feedback = self.compute_plan_cost(problem,
                                                                      result)
                                else:
                                    feedback = self.get_feedback(engine,
                                                                 metric,
                                                                 result)
                            else:
                                feedback = self.get_feedback(engine,
                                                             metric,
                                                             result)
                        elif result.status == \
                                up.engines.PlanGenerationResultStatus.\
                                UNSOLVABLE_PROVEN:
                            if self.verbose:
                                print("Problem proven unsolvable.\n")
                            feedback = 'unsolvable'
                        elif result.status == \
                                up.engines.PlanGenerationResultStatus.\
                                INTERNAL_ERROR:
                            if self.verbose:
                                print("Internal Error.\n")
                            feedback = None
                        else:
                            if self.verbose:
                                print("No plan found.\n")
                            feedback = None
                    except Exception as e:
                        if self.verbose:
                            print("No plan found.\n")
                            print('Exception:', e)
                        feedback = None

        elif plantype == 'AnytimePlanner':
            config = self.transform_conf_from_ac(engine, config, plantype)
            parametrizable_engines(engine)
            if gb_listener is not None:
                with AnytimePlanner(name=engine,
                                    params=config,
                                    output_stream=gb_listener) as planner:
                    try:
                        planner = linux_timeout(planner, engine,
                                                timelimit, config)
                        results = planner.get_solutions(problem,
                                                        timeout=timelimit)
                        for res in results:
                            if res is not None and \
                                (res.status ==
                                    up.engines.PlanGenerationResultStatus.
                                    SOLVED_SATISFICING or res.status ==
                                    up.engines.PlanGenerationResultStatus.
                                    SOLVED_OPTIMALLY):
                                result = res
                                if self.verbose:
                                    print("Result found.\n")
                                feedback = self.compute_plan_cost(problem, res)
                                results.close()
                                break
                            elif res is not None and \
                                    res.status == \
                                    up.engines.PlanGenerationResultStatus.\
                                    UNSOLVABLE_PROVEN:
                                if self.verbose:
                                    print("Problem proven unsolvable.\n")
                                results.close()
                                feedback = 'unsolvable'
                                break
                            elif res is not None and \
                                res.status == \
                                    up.engines.PlanGenerationResultStatus.\
                                    INTERNAL_ERROR:
                                if self.verbose:
                                    print("Internal Error.\n")
                                results.close()
                                feedback = None
                                break
                            elif res is not None and \
                                res.status == \
                                    up.engines.PlanGenerationResultStatus.\
                                    INTERMEDIATE:
                                feedback = self.compute_plan_cost(problem, res)
                                results.close()
                                break
                            else:
                                if self.verbose:
                                    print("No plan found.\n")
                                results.close()
                                feedback = None
                                break
                    except Exception as e:
                        if self.verbose:
                            print("No plan found.\n")
                            print('Exception:', e)
                        try:
                            results.close()
                        except Exception:
                            pass
                        feedback = None
            else:
                with AnytimePlanner(name=engine,
                                    params=config) as planner:
                    try:
                        planner = linux_timeout(planner, engine,
                                                timelimit, config)
                        results = planner.get_solutions(problem,
                                                        timeout=timelimit)
                        for res in results:
                            if res is not None and \
                                (res.status ==
                                    up.engines.PlanGenerationResultStatus.
                                    SOLVED_SATISFICING or res.status ==
                                    up.engines.PlanGenerationResultStatus.
                                    SOLVED_OPTIMALLY):
                                if self.verbose:
                                    print("Result found.\n")
                                feedback = self.compute_plan_cost(problem, res)
                                results.close()
                                break
                            elif res is not None and \
                                    res.status == \
                                    up.engines.PlanGenerationResultStatus.\
                                    UNSOLVABLE_PROVEN:
                                if self.verbose:
                                    print("Problem proven unsolvable.\n")
                                feedback = 'unsolvable'
                                results.close()
                                break
                            elif res is not None and \
                                res.status == \
                                    up.engines.PlanGenerationResultStatus.\
                                    INTERNAL_ERROR:
                                if self.verbose:
                                    print("Internal Error.\n")
                                feedback = None
                                results.close()
                                break
                            elif res is not None and \
                                res.status == \
                                    up.engines.PlanGenerationResultStatus.\
                                    INTERMEDIATE:
                                feedback = self.compute_plan_cost(problem, res)
                                results.close()
                                break
                            else:
                                if self.verbose:
                                    print("No plan found.\n")
                                feedback = None
                                results.close()
                                break
                    except Exception as e:
                        if self.verbose:
                            print("No plan found.\n")
                            print('Exception:', e)
                        feedback = None
                        try:
                            results.close()
                        except Exception:
                            pass

        return feedback



"""Functionalities for managing and calling SMAC."""
from smac import Scenario
from smac import AlgorithmConfigurationFacade
from unified_planning.exceptions import UPProblemDefinitionError, UPException
import os
import dill
import sys

import timeit
import signal
from contextlib import contextmanager
from pebble import concurrent
from pebble.common import ProcessExpired

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from AC_interface import *
from configurators import Configurator
from utils.patch_daskrunner import patch_daskrunner

# We do this only to be able to set the SMAC output dir!
AlgorithmConfigurationFacade = patch_daskrunner(AlgorithmConfigurationFacade)


class TimeoutException(Exception):
    pass


@contextmanager
def time_limit(seconds):
    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class SmacConfigurator(Configurator):
    """Configurator functions."""

    def __init__(self):
        """Initialize Smac configurator."""
        Configurator.__init__(self)
        self.crash_cost = 0
        self.planner_timelimit = 0
        self.engine = None
        self.gaci = None

    def get_feedback_function(self, gaci, engine, metric, mode,
                              gray_box=False):
        """
        Generate a function to run the planning engine and obtain feedback.

        :param ACInterface gaci: Algorithm Configuration Interface object.
        :param str engine: Name of the planning engine.
        :param str metric: Metric to optimize, either 'runtime' or 'quality'.
        :param str mode: Type of planning.
        :param bool gray_box: True if using a gray box, False otherwise (optional).

        :return: A planner feedback function that takes configuration, instance, seed, and reader.
        :rtype: function

        :raises ValueError: If the provided engine is not supported for the given metric and mode.
        """

        if engine in self.capabilities[metric][mode]:
            self.metric = metric

            def planner_feedback(config, instance, seed, reader):
                start = timeit.default_timer()
                instance_p = f'{instance}'
                # Since Smac handles time limits itself,
                # we do not use concurrent, as with other AC tools
                if metric == 'quality':
                    timelimit = self.planner_timelimit - self.patience
                else:
                    timelimit = self.planner_timelimit

                try:
                    if isinstance(self.train_set, dict) and \
                            isinstance(self.train_set[instance_p], tuple):
                        domain = self.train_set[instance_p][1]
                        problem = self.train_set[instance_p][0]
                        pddl_problem = reader.parse_problem(domain, problem)
                    else:
                        domain_path = instance_p.rsplit('/', 1)[0]
                        domain = f'{domain_path}/domain.pddl'
                        pddl_problem = reader.parse_problem(f'{domain}',
                                                            f'{instance_p}')

                    if engine == 'tamer' or engine == 'pyperplan':
                        @concurrent.process(timeout=timelimit)
                        def solve(config, metric, engine,
                                  mode, pddl_problem):
                            feedback = \
                                gaci.run_engine_config(config,
                                                       metric, engine,
                                                       mode, pddl_problem,
                                                       timelimit)

                            return feedback
                    else:
                        def solve(config, metric, engine,
                                  mode, pddl_problem):
                            feedback = \
                                gaci.run_engine_config(config,
                                                       metric, engine,
                                                       mode, pddl_problem,
                                                       timelimit)

                            return feedback

                    try:
                        if engine in ('tamer', 'pyperplan'):
                            future = solve(config, metric, engine,
                                           mode, pddl_problem)
                            try:
                                feedback = future.result()
                            except (TimeoutError, ProcessExpired) as err:
                                print(err)
                                if metric == 'runtime':
                                    feedback = timelimit
                                elif metric == 'quality':
                                    feedback = self.crash_cost

                        else:
                            with time_limit(timelimit):
                                feedback = solve(config, metric, engine,
                                                 mode, pddl_problem)

                    except TimeoutException:
                        if metric == 'runtime':
                            feedback = timelimit
                        elif metric == 'quality':
                            feedback = self.crash_cost

                except (AssertionError, NotImplementedError,
                        UPProblemDefinitionError, UPException,
                        UnicodeDecodeError) as err:
                    if self.verbose:
                        print('\n** Error in planning engine!', err)
                    if metric == 'runtime':
                        feedback = timelimit
                    elif metric == 'quality':
                        feedback = self.crash_cost

                if feedback == 'unsolvable':
                    if metric == 'runtime':
                        feedback = timelimit
                    elif metric == 'quality':
                        feedback = self.crash_cost
              
                if feedback is not None:
                    # SMAC always minimizes
                    if metric == 'quality':
                        self.print_feedback(engine, instance, feedback)
                        return feedback
                    # Solving runtime optimization by passing
                    # runtime as result, since smac minimizes it
                    elif metric == 'runtime':
                        if engine in ('tamer', 'pyperplan'):
                            feedback = timeit.default_timer() - start
                            if feedback > timelimit:
                                feedback = timelimit
                            self.print_feedback(engine, instance, feedback)
                        else:
                            if feedback > timelimit:
                                feedback = timelimit
                            self.print_feedback(engine, instance, feedback)
                        return feedback
                else:
                    # Penalizing failed runs
                    if metric == 'runtime':
                        # Penalty is max runtime in runtime scenario
                        feedback = timelimit
                        self.print_feedback(engine, instance, feedback)
                    elif metric == 'quality':
                        # Penalty is defined by user in quality scenario
                        feedback = self.crash_cost
                        self.print_feedback(engine, instance, feedback)

                    return feedback

            try:
                import up_ac
                path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
                path += '/utils'
            except ImportError:
                path = os.getcwd().rsplit('up_ac', 1)[0]
                if path[-1] != "/":
                    path += "/"
                path += 'up_ac/utils'

            self.feedback_path = path

            dill.dump(
                planner_feedback, open(
                    f'{path}/feedback.pkl', 'wb'),
                recurse=True)

            return planner_feedback
        else:
            if self.verbose:
                print(f'Algorithm Configuration for {metric} of {engine}' + \
                      f' in {mode} is not supported.')
            return None

    def set_scenario(self, engine, param_space, gaci,
                     configuration_time=120, n_trials=400, min_budget=1,
                     max_budget=3, crash_cost=10000, planner_timelimit=30,
                     n_workers=1, instances=[], instance_features=None,
                     output_dir='smac3_output', metric='runtime',
                     patience=10):
        """
        Set up the algorithm configuration scenario for SMAC (Sequential Model-based Algorithm Configuration).

        :param str engine: The name of the planning engine.
        :param ConfigSpace.ConfigurationSpace param_space: ConfigSpace object defining the parameter space.
        :param ACInterface gaci: AC interface object.
        :param int configuration_time: Overall configuration time budget in seconds (default is 120).
        :param int n_trials: Maximum number of engine evaluations (default is 400).
        :param int min_budget: Minimum number of instances to use (default is 1).
        :param int max_budget: Maximum number of instances to use (default is 3).
        :param int crash_cost: The cost to use if the engine fails (default is 0).
        :param int planner_timelimit: Maximum runtime per evaluation for the planner (default is 30).
        :param int n_workers: Number of cores to utilize (default is 1).
        :param list instances: List of problem instance paths (default is empty list, uses train_set).
        :param dict instance_features: Dictionary containing instance names and lists of features (default is None).
        :param str metric: The optimization metric, either 'runtime' or 'quality' (default is 'runtime').

        :raises ValueError: If an unsupported metric is provided.
        """
        if not instances:
            instances = self.train_set
        if isinstance(instances[0], tuple):
            train_set_dict = {}
            clean_inst = instances
            instances = []
            for inst in clean_inst:
                train_set_dict[inst[0]] = inst
                instances.append(inst[0])
            self.train_set = train_set_dict
        if metric == 'runtime':
            if crash_cost != planner_timelimit:
                crash_cost = planner_timelimit
        self.crash_cost = crash_cost
        self.patience = patience
        self.metric = metric
        if self.metric == 'quality':
            self.planner_timelimit = planner_timelimit + patience
        else:
            self.planner_timelimit = planner_timelimit
        self.engine = engine
        self.gaci = gaci
        self.n_workers = n_workers
        self.scenarios = {}

        for n in range(self.n_workers):
            scenario_dict = {'configspace': param_space,
                             'walltime_limit': configuration_time,
                             'n_trials': n_trials,
                             'min_budget': min_budget,
                             'max_budget': max_budget,
                             'deterministic': True,
                             'crash_cost': crash_cost,  
                             'trial_walltime_limit': planner_timelimit,  
                             'use_default_config': True,
                             'n_workers': 1,
                             'instances': instances,
                             'instance_features': instance_features}

            od = output_dir + f'_{n}'

            self.scenarios[n] = Scenario(**scenario_dict,
                                         output_directory=od)

        if self.verbose:
            print('\nSMAC scenario is set.\n')

    def optimize(self, feedback_function=None, gray_box=False):
        """
        Run the algorithm configuration optimization.

        :param function feedback_function: A function to run the engine and obtain feedback (optional).
        :param bool gray_box: True if gray box usage is enabled, False otherwise (optional).

        :return: A tuple containing the best configuration found and additional information (if available).
                Returns None if feedback_function is not provided.
        :rtype: tuple or None
        """
        if feedback_function is not None:
            # Import feedback function, since dask cannot pickle local objects            
            try:
                import up_ac
                path = \
                    '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
                path += '/utils'
            except ImportError:
                path = os.getcwd().rsplit('up_ac', 1)[0]
                if path[-1] != "/":
                    path += "/"
                path += 'up_ac/utils'
            sys.path.append(r"{}".format(path))
            from load_smac_feedback import get_feedback

            if self.verbose:
                print('\nStarting Parameter optimization\n')

            ac_dict = {}
            self.incumbent_dict = {}
            self.incumbent_dicts_dict = {}
            self.runhistories = {}
            self.costs = []
            threads = []
            results = [None for n in range(self.n_workers)]

            def optimize(smac, results, n):
                results[n] = smac.optimize()

            import threading

            for n in range(self.n_workers):

                print('\n\n\n')
                print('!!!!!!!!!!!!!!')
                print('Running AC Nr.', n)
                print('!!!!!!!!!!!!!!')
                print('\n\n\n')
 
                ac_dict[n] = AlgorithmConfigurationFacade(
                    self.scenarios[n],
                    get_feedback,
                    overwrite=True)

                threads.append(threading.Thread(target=optimize,
                                                args=(ac_dict[n], results, n)))

            for n in range(self.n_workers):
                threads[n].start()

            for n in range(self.n_workers):
                threads[n].join()

            for n in range(self.n_workers):
                self.incumbent_dict[n] = results[n]
                if results[n] is not None:
                    self.incumbent_dicts_dict[n] = results[n].get_dictionary()
                    self.runhistories[n] = ac_dict[n]._runhistory
                    self.costs.append(
                        self.runhistories[n].get_cost(self.incumbent_dict[n]))
                else:
                    self.incumbent_dicts_dict[n] = None
                    self.runhistories[n] = None
                    if self.metric == 'quality':
                        self.costs.append(self.crash_cost)
                    else:
                        self.costs.append(self.planner_timelimit)

            for n, incumbent in self.incumbent_dicts_dict.items():
                print('Incumbent', n, incumbent, 'has cost', self.costs[n])

            index_best = \
                min(range(len(self.costs)), key=self.costs.__getitem__)
            self.incumbent = self.incumbent_dicts_dict[index_best]

            if self.verbose:
                print('\nBest Configuration found is:\n',
                      self.incumbent)

            return self.incumbent, None
        else:
            return None, None
